@@ -4,6 +4,7 @@ use Mojo::Base 'Mojo::EventEmitter';
 # "Fry: Since when is the Internet about robbing people of their privacy?
 #  Bender: August 6, 1991."
 use Mojo::IOLoop;
+use Mojo::Promise;
 use Mojo::Util qw(monkey_patch term_escape);
 use Mojo::UserAgent::CookieJar;
 use Mojo::UserAgent::Proxy;
@@ -17,24 +18,27 @@ has ca              => sub { $ENV{MOJO_CA_FILE} };
 has cert            => sub { $ENV{MOJO_CERT_FILE} };
 has connect_timeout => sub { $ENV{MOJO_CONNECT_TIMEOUT} || 10 };
 has cookie_jar      => sub { Mojo::UserAgent::CookieJar->new };
-has 'local_address';
+has [qw(local_address max_response_size)];
 has inactivity_timeout => sub { $ENV{MOJO_INACTIVITY_TIMEOUT} // 20 };
 has ioloop             => sub { Mojo::IOLoop->new };
 has key                => sub { $ENV{MOJO_KEY_FILE} };
 has max_connections    => 5;
-has max_redirects => sub { $ENV{MOJO_MAX_REDIRECTS} || 0 };
-has proxy => sub { Mojo::UserAgent::Proxy->new };
-has request_timeout => sub { $ENV{MOJO_REQUEST_TIMEOUT} // 0 };
-has server => sub { Mojo::UserAgent::Server->new(ioloop => shift->ioloop) };
+has max_redirects      => sub { $ENV{MOJO_MAX_REDIRECTS} || 0 };
+has proxy              => sub { Mojo::UserAgent::Proxy->new };
+has request_timeout    => sub { $ENV{MOJO_REQUEST_TIMEOUT} // 0 };
+has server     => sub { Mojo::UserAgent::Server->new(ioloop => shift->ioloop) };
 has transactor => sub { Mojo::UserAgent::Transactor->new };
 
 # Common HTTP methods
 for my $name (qw(DELETE GET HEAD OPTIONS PATCH POST PUT)) {
   monkey_patch __PACKAGE__, lc $name, sub {
-    my $self = shift;
-    my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+    my ($self, $cb) = (shift, ref $_[-1] eq 'CODE' ? pop : undef);
     return $self->start($self->build_tx($name, @_), $cb);
   };
+  monkey_patch __PACKAGE__, lc($name) . '_p', sub {
+    my $self = shift;
+    return $self->start_p($self->build_tx($name, @_));
+  };
 }
 
 sub DESTROY { Mojo::Util::_global_destruction() or shift->_cleanup }
@@ -51,46 +55,61 @@ sub start {
   # Non-blocking
   if ($cb) {
     warn "-- Non-blocking request (@{[_url($tx)]})\n" if DEBUG;
-    return $self->_start(1, $tx, $cb);
+    return $self->_start(Mojo::IOLoop->singleton, $tx, $cb);
   }
 
   # Blocking
   warn "-- Blocking request (@{[_url($tx)]})\n" if DEBUG;
-  $self->_start(0, $tx => sub { shift->ioloop->stop; $tx = shift });
+  $self->_start($self->ioloop, $tx => sub { shift->ioloop->stop; $tx = shift });
   $self->ioloop->start;
 
   return $tx;
 }
 
+sub start_p {
+  my ($self, $tx) = @_;
+
+  my $promise = Mojo::Promise->new;
+  $self->start(
+    $tx => sub {
+      my ($self, $tx) = @_;
+      my $err = $tx->error;
+      return $promise->reject($err->{message}) if $err && !$err->{code};
+      return $promise->reject('WebSocket handshake failed')
+        if $tx->req->is_handshake && !$tx->is_websocket;
+      $promise->resolve($tx);
+    }
+  );
+
+  return $promise;
+}
+
 sub websocket {
-  my $self = shift;
-  my $cb = ref $_[-1] eq 'CODE' ? pop : undef;
+  my ($self, $cb) = (shift, pop);
   $self->start($self->build_websocket_tx(@_), $cb);
 }
 
+sub websocket_p {
+  my $self = shift;
+  return $self->start_p($self->build_websocket_tx(@_));
+}
+
 sub _cleanup {
   my $self = shift;
-
-  # Clean up active connections (by closing them)
   delete $self->{pid};
   $self->_finish($_, 1) for keys %{$self->{connections} || {}};
-
-  # Clean up keep-alive connections
-  my $loop = $self->_loop(0);
-  $loop->remove($_->[1]) for @{delete $self->{queue} || []};
-  $loop = $self->_loop(1);
-  $loop->remove($_->[1]) for @{delete $self->{nb_queue} || []};
-
   return $self;
 }
 
 sub _connect {
-  my ($self, $nb, $peer, $tx, $handle, $cb) = @_;
+  my ($self, $loop, $peer, $tx, $handle, $cb) = @_;
 
   my $t = $self->transactor;
   my ($proto, $host, $port) = $peer ? $t->peer($tx) : $t->endpoint($tx);
-  my %options
-    = (address => $host, port => $port, timeout => $self->connect_timeout);
+
+  my %options = (timeout => $self->connect_timeout);
+  if ($proto eq 'http+unix') { $options{path} = $host }
+  else                       { @options{qw(address port)} = ($host, $port) }
   if (my $local = $self->local_address) { $options{local_address} = $local }
   $options{handle} = $handle if $handle;
 
@@ -98,9 +117,7 @@ sub _connect {
   if ($proto eq 'socks') {
     @options{qw(socks_address socks_port)} = @options{qw(address port)};
     ($proto, @options{qw(address port)}) = $t->endpoint($tx);
-    my $req      = $tx->req;
-    my $userinfo = $req->proxy->userinfo;
-    $req->proxy(0);
+    my $userinfo = $tx->req->via_proxy(0)->proxy->userinfo;
     @options{qw(socks_user socks_pass)} = split ':', $userinfo if $userinfo;
   }
 
@@ -110,7 +127,7 @@ sub _connect {
 
   weaken $self;
   my $id;
-  return $id = $self->_loop($nb)->client(
+  return $id = $loop->client(
     %options => sub {
       my ($loop, $err, $stream) = @_;
 
@@ -129,32 +146,33 @@ sub _connect {
 }
 
 sub _connect_proxy {
-  my ($self, $nb, $old, $cb) = @_;
+  my ($self, $loop, $old, $cb) = @_;
 
   # Start CONNECT request
   return undef unless my $new = $self->transactor->proxy_connect($old);
   return $self->_start(
-    ($nb, $new) => sub {
+    ($loop, $new) => sub {
       my ($self, $tx) = @_;
 
-      # CONNECT failed (connection needs to be kept alive)
-      $old->res->error({message => 'Proxy connection failed'})
-        and return $self->$cb($old)
-        if $tx->error || !$tx->res->is_status_class(200) || !$tx->keep_alive;
-
-      # Start real transaction
-      $old->req->proxy(0);
+      # CONNECT failed
+      $old->previous($tx)->req->via_proxy(0);
       my $id = $tx->connection;
-      return $self->_start($nb, $old->connection($id), $cb)
+      if ($tx->error || !$tx->res->is_success || !$tx->keep_alive) {
+        $old->res->error({message => 'Proxy connection failed'});
+        $self->_remove($id) if $id;
+        return $self->$cb($old);
+      }
+
+      # Start real transaction without TLS upgrade
+      return $self->_start($loop, $old->connection($id), $cb)
         unless $tx->req->url->protocol eq 'https';
 
-      # TLS upgrade
-      my $loop   = $self->_loop($nb);
+      # TLS upgrade before starting the real transaction
       my $handle = $loop->stream($id)->steal_handle;
-      $loop->remove($id);
-      $id = $self->_connect($nb, 0, $old, $handle,
-        sub { shift->_start($nb, $old->connection($id), $cb) });
-      $self->{connections}{$id} = {cb => $cb, nb => $nb, tx => $old};
+      $self->_remove($id);
+      $id = $self->_connect($loop, 0, $old, $handle,
+        sub { shift->_start($loop, $old->connection($id), $cb) });
+      $self->{connections}{$id} = {cb => $cb, ioloop => $loop, tx => $old};
     }
   );
 }
@@ -162,53 +180,52 @@ sub _connect_proxy {
 sub _connected {
   my ($self, $id) = @_;
 
-  # Inactivity timeout
-  my $c = $self->{connections}{$id};
-  my $stream
-    = $self->_loop($c->{nb})->stream($id)->timeout($self->inactivity_timeout);
-
-  # Store connection information in transaction
+  my $c      = $self->{connections}{$id};
+  my $stream = $c->{ioloop}->stream($id)->timeout($self->inactivity_timeout);
   my $tx     = $c->{tx}->connection($id);
   my $handle = $stream->handle;
-  $tx->local_address($handle->sockhost)->local_port($handle->sockport);
-  $tx->remote_address($handle->peerhost)->remote_port($handle->peerport);
+  unless ($handle->isa('IO::Socket::UNIX')) {
+    $tx->local_address($handle->sockhost)->local_port($handle->sockport);
+    $tx->remote_address($handle->peerhost)->remote_port($handle->peerport);
+  }
 
-  # Start writing
   weaken $self;
   $tx->on(resume => sub { $self->_write($id) });
   $self->_write($id);
 }
 
 sub _connection {
-  my ($self, $nb, $tx, $cb) = @_;
+  my ($self, $loop, $tx, $cb) = @_;
 
   # Reuse connection
   my ($proto, $host, $port) = $self->transactor->endpoint($tx);
-  my $id = $tx->connection || $self->_dequeue($nb, "$proto:$host:$port", 1);
+  my $id = $tx->connection || $self->_dequeue($loop, "$proto:$host:$port", 1);
   if ($id) {
     warn "-- Reusing connection $id ($proto://$host:$port)\n" if DEBUG;
-    $self->{connections}{$id} = {cb => $cb, nb => $nb, tx => $tx};
+    @{$self->{connections}{$id}}{qw(cb tx)} = ($cb, $tx);
     $tx->kept_alive(1) unless $tx->connection;
     $self->_connected($id);
     return $id;
   }
 
   # CONNECT request to proxy required
-  if (my $id = $self->_connect_proxy($nb, $tx, $cb)) { return $id }
+  if (my $id = $self->_connect_proxy($loop, $tx, $cb)) { return $id }
 
-  # Connect
-  $id = $self->_connect($nb, 1, $tx, undef, \&_connected);
+  # New connection
+  $tx->res->error({message => "Unsupported protocol: $proto"})
+    and return $loop->next_tick(sub { $self->$cb($tx) })
+    unless $proto eq 'http' || $proto eq 'https' || $proto eq 'http+unix';
+  $id = $self->_connect($loop, 1, $tx, undef, \&_connected);
   warn "-- Connect $id ($proto://$host:$port)\n" if DEBUG;
-  $self->{connections}{$id} = {cb => $cb, nb => $nb, tx => $tx};
+  $self->{connections}{$id} = {cb => $cb, ioloop => $loop, tx => $tx};
 
   return $id;
 }
 
 sub _dequeue {
-  my ($self, $nb, $name, $test) = @_;
+  my ($self, $loop, $name, $test) = @_;
 
-  my $loop = $self->_loop($nb);
-  my $old = $self->{$nb ? 'nb_queue' : 'queue'} ||= [];
+  my $old = $self->{queue}{$loop} ||= [];
   my ($found, @new);
   for my $queued (@$old) {
     push @new, $queued and next if $found || !grep { $_ eq $name } @$queued;
@@ -222,16 +239,6 @@ sub _dequeue {
   return $found;
 }
 
-sub _enqueue {
-  my ($self, $nb, $name, $id) = @_;
-
-  # Enforce connection limit
-  my $queue = $self->{$nb ? 'nb_queue' : 'queue'} ||= [];
-  my $max = $self->max_connections;
-  $self->_remove(shift(@$queue)->[1]) while @$queue && @$queue >= $max;
-  $max ? push @$queue, [$name, $id] : $self->_loop($nb)->stream($id)->close;
-}
-
 sub _error {
   my ($self, $id, $err) = @_;
   my $tx = $self->{connections}{$id}{tx};
@@ -242,16 +249,19 @@ sub _error {
 sub _finish {
   my ($self, $id, $close) = @_;
 
-  # Remove request timeout
+  # Remove request timeout and finish transaction
   return unless my $c = $self->{connections}{$id};
-  my $loop = $self->_loop($c->{nb});
-  $loop->remove($c->{timeout}) if $c->{timeout};
+  $c->{ioloop}->remove($c->{timeout}) if $c->{timeout};
+  return $self->_reuse($id, $close) unless my $old = $c->{tx};
 
-  return $self->_remove($id, $close) unless my $old = $c->{tx};
-  $old->client_close($close);
+  # Premature connection close
+  my $res = $old->closed->res->finish;
+  if ($close && !$res->code && !$res->error) {
+    $res->error({message => 'Premature connection close'});
+  }
 
-  # Finish WebSocket
-  return $self->_remove($id, 1) if $old->is_websocket;
+  # Always remove connection for WebSockets
+  return $self->_remove($id) if $old->is_websocket;
 
   $self->cookie_jar->collect($old);
 
@@ -263,66 +273,74 @@ sub _finish {
     return $new->client_read($old->res->content->leftovers);
   }
 
-  # Finish normal connection and handle redirects
-  $self->_remove($id, $close);
+  # CONNECT requests always have a follow-up request
+  $self->_reuse($id, $close) unless uc $old->req->method eq 'CONNECT';
+  $res->error({message => $res->message, code => $res->code}) if $res->is_error;
   $c->{cb}($self, $old) unless $self->_redirect($c, $old);
 }
 
-sub _loop { $_[1] ? Mojo::IOLoop->singleton : $_[0]->ioloop }
-
 sub _read {
   my ($self, $id, $chunk) = @_;
 
   # Corrupted connection
-  return                     unless my $c  = $self->{connections}{$id};
-  return $self->_remove($id) unless my $tx = $c->{tx};
+  return $self->_remove($id) unless my $tx = $self->{connections}{$id}{tx};
 
-  # Process incoming data
   warn term_escape "-- Client <<< Server (@{[_url($tx)]})\n$chunk\n" if DEBUG;
   $tx->client_read($chunk);
-  if    ($tx->is_finished) { $self->_finish($id) }
-  elsif ($tx->is_writing)  { $self->_write($id) }
+  $self->_finish($id) if $tx->is_finished;
 }
 
 sub _redirect {
   my ($self, $c, $old) = @_;
   return undef unless my $new = $self->transactor->redirect($old);
   return undef unless @{$old->redirects} < $self->max_redirects;
-  return $self->_start($c->{nb}, $new, delete $c->{cb});
+  return $self->_start($c->{ioloop}, $new, delete $c->{cb});
 }
 
 sub _remove {
+  my ($self, $id) = @_;
+  my $c = delete $self->{connections}{$id};
+  $self->_dequeue($c->{ioloop}, $id);
+  $c->{ioloop}->remove($id);
+}
+
+sub _reuse {
   my ($self, $id, $close) = @_;
 
-  # Close connection
-  my $c = delete $self->{connections}{$id} || {};
-  my $tx = $c->{tx};
-  return map { $self->_dequeue($_, $id); $self->_loop($_)->remove($id) } 1, 0
-    if $close || !$tx || !$tx->keep_alive || $tx->error;
+  # Connection close
+  my $c   = $self->{connections}{$id};
+  my $tx  = delete $c->{tx};
+  my $max = $self->max_connections;
+  return $self->_remove($id)
+    if $close || !$tx || !$max || !$tx->keep_alive || $tx->error;
 
-  # Keep connection alive (CONNECT requests get upgraded)
-  $self->_enqueue($c->{nb}, join(':', $self->transactor->endpoint($tx)), $id)
-    unless uc $tx->req->method eq 'CONNECT';
+  # Keep connection alive
+  my $queue = $self->{queue}{$c->{ioloop}} ||= [];
+  $self->_remove(shift(@$queue)->[1]) while @$queue && @$queue >= $max;
+  push @$queue, [join(':', $self->transactor->endpoint($tx)), $id];
 }
 
 sub _start {
-  my ($self, $nb, $tx, $cb) = @_;
+  my ($self, $loop, $tx, $cb) = @_;
 
-  # Application server
+  # Application serve
   my $url = $tx->req->url;
   unless ($url->is_abs) {
-    my $base = $nb ? $self->server->nb_url : $self->server->url;
-    $url->scheme($base->scheme)->authority($base->authority);
+    my $base
+      = $loop == $self->ioloop ? $self->server->url : $self->server->nb_url;
+    $url->scheme($base->scheme)->host($base->host)->port($base->port);
   }
 
-  $_ && $_->prepare($tx) for $self->proxy, $self->cookie_jar;
+  $_->prepare($tx) for $self->proxy, $self->cookie_jar;
+  my $max = $self->max_response_size;
+  $tx->res->max_message_size($max) if defined $max;
 
-  # Connect and add request timeout if necessary
-  my $id = $self->emit(start => $tx)->_connection($nb, $tx, $cb);
+  $self->emit(start => $tx);
+  return undef unless my $id = $self->_connection($loop, $tx, $cb);
   if (my $timeout = $self->request_timeout) {
     weaken $self;
-    $self->{connections}{$id}{timeout} = $self->_loop($nb)
-      ->timer($timeout => sub { $self->_error($id, 'Request timeout') });
+    $self->{connections}{$id}{timeout}
+      = $loop->timer($timeout => sub { $self->_error($id, 'Request timeout') });
   }
 
   return $id;
@@ -333,20 +351,15 @@ sub _url { shift->req->url->to_abs }
 sub _write {
   my ($self, $id) = @_;
 
-  # Get and write chunk
-  return unless my $c  = $self->{connections}{$id};
-  return unless my $tx = $c->{tx};
-  return if !$tx->is_writing || $c->{writing}++;
+  # Protect from resume event recursion
+  my $c = $self->{connections}{$id};
+  return if !(my $tx = $c->{tx}) || $c->{writing};
+  local $c->{writing} = 1;
   my $chunk = $tx->client_write;
-  delete $c->{writing};
   warn term_escape "-- Client >>> Server (@{[_url($tx)]})\n$chunk\n" if DEBUG;
-  my $stream = $self->_loop($c->{nb})->stream($id)->write($chunk);
-  $self->_finish($id) if $tx->is_finished;
-
-  # Continue writing
-  return unless $tx->is_writing;
+  return unless length $chunk;
   weaken $self;
-  $stream->write('' => sub { $self->_write($id) });
+  $c->{ioloop}->stream($id)->write($chunk => sub { $self->_write($id) });
 }
 
 1;
@@ -361,31 +374,29 @@ Mojo::UserAgent - Non-blocking I/O HTTP and WebSocket user agent
 
   use Mojo::UserAgent;
 
-  # Say hello to the Unicode snowman and include an Accept header
-  my $ua = Mojo::UserAgent->new;
-  say $ua->get('www.☃.net?hello=there' => {Accept => '*/*'})->res->body;
+  # Fine grained response handling (dies on connection errors)
+  my $ua  = Mojo::UserAgent->new;
+  my $res = $ua->get('mojolicious.org/perldoc')->result;
+  if    ($res->is_success)  { say $res->body }
+  elsif ($res->is_error)    { say $res->message }
+  elsif ($res->code == 301) { say $res->headers->location }
+  else                      { say 'Whatever...' }
 
-  # Form POST (application/x-www-form-urlencoded) with exception handling
-  my $tx = $ua->post('https://metacpan.org/search' => form => {q => 'mojo'});
-  if (my $res = $tx->success) { say $res->body }
-  else {
-    my $err = $tx->error;
-    die "$err->{code} response: $err->{message}" if $err->{code};
-    die "Connection error: $err->{message}";
-  }
+  # Say hello to the Unicode snowman and include an Accept header
+  say $ua->get('www.☃.net?hello=there' => {Accept => '*/*'})->result->body;
 
   # Extract data from HTML and XML resources with CSS selectors
-  say $ua->get('www.perl.org')->res->dom->at('title')->text;
+  say $ua->get('www.perl.org')->result->dom->at('title')->text;
 
   # Scrape the latest headlines from a news site
   say $ua->get('blogs.perl.org')
-    ->res->dom->find('h2 > a')->map('text')->join("\n");
+    ->result->dom->find('h2 > a')->map('text')->join("\n");
 
   # IPv6 PUT request with Content-Type header and content
   my $tx = $ua->put('[::1]:3000' => {'Content-Type' => 'text/plain'} => 'Hi!');
 
   # Quick JSON API request with Basic authentication
-  my $value = $ua->get('https://sri:s3cret@example.com/test.json')->res->json;
+  my $value = $ua->get('https://sri:t3st@example.com/test.json')->result->json;
 
   # JSON POST (application/json) with TLS certificate authentication
   my $tx = $ua->cert('tls.crt')->key('tls.key')
@@ -394,29 +405,43 @@ Mojo::UserAgent - Non-blocking I/O HTTP and WebSocket user agent
   # Search DuckDuckGo anonymously through Tor
   $ua->proxy->http('socks://127.0.0.1:9050');
   say $ua->get('api.3g2upl4pq6kufc4m.onion/?q=mojolicious&format=json')
-    ->res->json('/Abstract');
+    ->result->json('/Abstract');
+
+  # GET request via UNIX domain socket "/tmp/myapp.sock" (percent encoded slash)
+  say $ua->get('http+unix://%2Ftmp%2Fmyapp.sock/perldoc')->result->body;
 
   # Follow redirects to download Mojolicious from GitHub
   $ua->max_redirects(5)
     ->get('https://www.github.com/kraih/mojo/tarball/master')
-    ->res->content->asset->move_to('/home/sri/mojo.tar.gz');
+    ->result->content->asset->move_to('/home/sri/mojo.tar.gz');
 
-  # Non-blocking concurrent requests
-  Mojo::IOLoop->delay(
-    sub {
-      my $delay = shift;
-      $ua->get('mojolicio.us' => $delay->begin);
-      $ua->get('cpan.org'     => $delay->begin);
-    },
-    sub {
-      my ($delay, $mojo, $cpan) = @_;
-      say $mojo->res->dom->at('title')->text;
-      say $cpan->res->dom->at('title')->text;
-    }
-  )->wait;
+  # Form POST (application/x-www-form-urlencoded) with manual exception handling
+  my $tx = $ua->post('https://metacpan.org/search' => form => {q => 'mojo'});
+  if (my $res = $tx->success) { say $res->body }
+  else {
+    my $err = $tx->error;
+    die "$err->{code} response: $err->{message}" if $err->{code};
+    die "Connection error: $err->{message}";
+  }
 
-  # Non-blocking WebSocket connection sending and receiving JSON messages
-  $ua->websocket('ws://example.com/echo.json' => sub {
+  # Non-blocking request
+  $ua->get('mojolicious.org' => sub {
+    my ($ua, $tx) = @_;
+    say $tx->result->dom->at('title')->text;
+  });
+  Mojo::IOLoop->start unless Mojo::IOLoop->is_running;
+
+  # Concurrent non-blocking requests (synchronized with promises)
+  my $mojo = $ua->get_p('mojolicious.org');
+  my $cpan = $ua->get_p('cpan.org');
+  Mojo::Promise->all($mojo, $cpan)->then(sub {
+    my ($mojo, $cpan) = @_;
+    say $mojo->[0]->result->dom->at('title')->text;
+    say $cpan->[0]->result->dom->at('title')->text;
+  })->wait;
+
+  # WebSocket connection sending and receiving JSON via UNIX domain socket
+  $ua->websocket('ws+unix://%2Ftmp%2Fmyapp.sock/echo.json' => sub {
     my ($ua, $tx) = @_;
     say 'WebSocket handshake failed!' and return unless $tx->is_websocket;
     $tx->on(json => sub {
@@ -431,9 +456,9 @@ Mojo::UserAgent - Non-blocking I/O HTTP and WebSocket user agent
 =head1 DESCRIPTION
 
 L<Mojo::UserAgent> is a full featured non-blocking I/O HTTP and WebSocket user
-agent, with IPv6, TLS, SNI, IDNA, HTTP/SOCKS5 proxy, Comet (long polling),
-keep-alive, connection pooling, timeout, cookie, multipart, gzip compression
-and multiple event loop support.
+agent, with IPv6, TLS, SNI, IDNA, HTTP/SOCKS5 proxy, UNIX domain socket, Comet
+(long polling), Promises/A+, keep-alive, connection pooling, timeout, cookie,
+multipart, gzip compression and multiple event loop support.
 
 All connections will be reset automatically if a new process has been forked,
 this allows multiple processes to share the same L<Mojo::UserAgent> object
@@ -443,7 +468,7 @@ For better scalability (epoll, kqueue) and to provide non-blocking name
 resolution, SOCKS5 as well as TLS support, the optional modules L<EV> (4.0+),
 L<Net::DNS::Native> (0.15+), L<IO::Socket::Socks> (0.64+) and
 L<IO::Socket::SSL> (1.94+) will be used automatically if possible. Individual
-features can also be disabled with the C<MOJO_NO_NDN>, C<MOJO_NO_SOCKS> and
+features can also be disabled with the C<MOJO_NO_NNR>, C<MOJO_NO_SOCKS> and
 C<MOJO_NO_TLS> environment variables.
 
 See L<Mojolicious::Guides::Cookbook/"USER AGENT"> for more.
@@ -526,7 +551,7 @@ L<Mojo::UserAgent::CookieJar> object.
     Mojo::Cookie::Response->new(
       name   => 'foo',
       value  => 'bar',
-      domain => 'mojolicio.us',
+      domain => 'mojolicious.org',
       path   => '/perldoc'
     )
   );
@@ -582,6 +607,18 @@ Maximum number of redirects the user agent will follow before it fails,
 defaults to the value of the C<MOJO_MAX_REDIRECTS> environment variable or
 C<0>.
 
+=head2 max_response_size
+
+  my $max = $ua->max_response_size;
+  $ua     = $ua->max_response_size(16777216);
+
+Maximum response size in bytes, defaults to the value of
+L<Mojo::Message::Response/"max_message_size">. Setting the value to C<0> will
+allow responses of indefinite size. Note that increasing this value can also
+drastically increase memory usage, should you for example attempt to parse an
+excessively large response body with the methods L<Mojo::Message/"dom"> or
+L<Mojo::Message/"json">.
+
 =head2 proxy
 
   my $proxy = $ua->proxy;
@@ -592,12 +629,16 @@ Proxy manager, defaults to a L<Mojo::UserAgent::Proxy> object.
   # Detect proxy servers from environment
   $ua->proxy->detect;
 
-  # Manually configure HTTP proxy (using CONNECT for HTTPS)
+  # Manually configure HTTP proxy (using CONNECT for HTTPS/WebSockets)
   $ua->proxy->http('http://127.0.0.1:8080')->https('http://127.0.0.1:8080');
 
   # Manually configure Tor (SOCKS5)
   $ua->proxy->http('socks://127.0.0.1:9050')->https('socks://127.0.0.1:9050');
 
+  # Manually configure UNIX domain socket (using CONNECT for HTTPS/WebSockets)
+  $ua->proxy->http('http+unix://%2Ftmp%2Fproxy.sock')
+    ->https('http+unix://%2Ftmp%2Fproxy.sock');
+
 =head2 request_timeout
 
   my $timeout = $ua->request_timeout;
@@ -620,8 +661,13 @@ The timeout will reset for every followed redirect.
 Application server relative URLs will be processed with, defaults to a
 L<Mojo::UserAgent::Server> object.
 
-  # Introspect
-  say for @{$ua->server->app->secrets};
+  # Mock web service
+  $ua->server->app(Mojolicious->new);
+  $ua->server->app->routes->get('/time' => sub {
+    my $c = shift;
+    $c->render(json => {now => time});
+  });
+  my $time = $ua->get('/time')->result->json->{now};
 
   # Change log level
   $ua->server->app->log->level('fatal');
@@ -651,7 +697,7 @@ implements the following new ones.
 
   my $tx = $ua->build_tx(GET => 'example.com');
   my $tx = $ua->build_tx(
-    PUT => 'http://example.com' => {Accept => '*/*'} => 'Hi!');
+    PUT => 'http://example.com' => {Accept => '*/*'} => 'Content!');
   my $tx = $ua->build_tx(
     PUT => 'http://example.com' => {Accept => '*/*'} => form => {a => 'b'});
   my $tx = $ua->build_tx(
@@ -706,7 +752,7 @@ L<Mojo::UserAgent::Transactor/"websocket">.
 =head2 delete
 
   my $tx = $ua->delete('example.com');
-  my $tx = $ua->delete('http://example.com' => {Accept => '*/*'} => 'Hi!');
+  my $tx = $ua->delete('http://example.com' => {Accept => '*/*'} => 'Content!');
   my $tx = $ua->delete(
     'http://example.com' => {Accept => '*/*'} => form => {a => 'b'});
   my $tx = $ua->delete(
@@ -719,14 +765,29 @@ implied). You can also append a callback to perform requests non-blocking.
 
   $ua->delete('http://example.com' => json => {a => 'b'} => sub {
     my ($ua, $tx) = @_;
-    say $tx->res->body;
+    say $tx->result->body;
   });
   Mojo::IOLoop->start unless Mojo::IOLoop->is_running;
 
+=head2 delete_p
+
+  my $promise = $ua->delete_p('http://example.com');
+
+Same as L</"delete">, but performs all requests non-blocking and returns a
+L<Mojo::Promise> object instead of accepting a callback.
+
+  $ua->delete_p('http://example.com' => json => {a => 'b'})->then(sub {
+    my $tx = shift;
+    say $tx->result->body;
+  })->catch(sub {
+    my $err = shift;
+    warn "Connection error: $err";
+  })->wait;
+
 =head2 get
 
   my $tx = $ua->get('example.com');
-  my $tx = $ua->get('http://example.com' => {Accept => '*/*'} => 'Hi!');
+  my $tx = $ua->get('http://example.com' => {Accept => '*/*'} => 'Content!');
   my $tx = $ua->get(
     'http://example.com' => {Accept => '*/*'} => form => {a => 'b'});
   my $tx = $ua->get(
@@ -739,14 +800,29 @@ perform requests non-blocking.
 
   $ua->get('http://example.com' => json => {a => 'b'} => sub {
     my ($ua, $tx) = @_;
-    say $tx->res->body;
+    say $tx->result->body;
   });
   Mojo::IOLoop->start unless Mojo::IOLoop->is_running;
 
+=head2 get_p
+
+  my $promise = $ua->get_p('http://example.com');
+
+Same as L</"get">, but performs all requests non-blocking and returns a
+L<Mojo::Promise> object instead of accepting a callback.
+
+  $ua->get_p('http://example.com' => json => {a => 'b'})->then(sub {
+    my $tx = shift;
+    say $tx->result->body;
+  })->catch(sub {
+    my $err = shift;
+    warn "Connection error: $err";
+  })->wait;
+
 =head2 head
 
   my $tx = $ua->head('example.com');
-  my $tx = $ua->head('http://example.com' => {Accept => '*/*'} => 'Hi!');
+  my $tx = $ua->head('http://example.com' => {Accept => '*/*'} => 'Content!');
   my $tx = $ua->head(
     'http://example.com' => {Accept => '*/*'} => form => {a => 'b'});
   my $tx = $ua->head(
@@ -759,14 +835,29 @@ implied). You can also append a callback to perform requests non-blocking.
 
   $ua->head('http://example.com' => json => {a => 'b'} => sub {
     my ($ua, $tx) = @_;
-    say $tx->res->body;
+    say $tx->result->body;
   });
   Mojo::IOLoop->start unless Mojo::IOLoop->is_running;
 
+=head2 head_p
+
+  my $promise = $ua->head_p('http://example.com');
+
+Same as L</"head">, but performs all requests non-blocking and returns a
+L<Mojo::Promise> object instead of accepting a callback.
+
+  $ua->head_p('http://example.com' => json => {a => 'b'})->then(sub {
+    my $tx = shift;
+    say $tx->result->body;
+  })->catch(sub {
+    my $err = shift;
+    warn "Connection error: $err";
+  })->wait;
+
 =head2 options
 
   my $tx = $ua->options('example.com');
-  my $tx = $ua->options('http://example.com' => {Accept => '*/*'} => 'Hi!');
+  my $tx = $ua->options('http://example.com' => {Accept => '*/*'} => 'Content!');
   my $tx = $ua->options(
     'http://example.com' => {Accept => '*/*'} => form => {a => 'b'});
   my $tx = $ua->options(
@@ -779,14 +870,29 @@ implied). You can also append a callback to perform requests non-blocking.
 
   $ua->options('http://example.com' => json => {a => 'b'} => sub {
     my ($ua, $tx) = @_;
-    say $tx->res->body;
+    say $tx->result->body;
   });
   Mojo::IOLoop->start unless Mojo::IOLoop->is_running;
 
+=head2 options_p
+
+  my $promise = $ua->options_p('http://example.com');
+
+Same as L</"options">, but performs all requests non-blocking and returns a
+L<Mojo::Promise> object instead of accepting a callback.
+
+  $ua->options_p('http://example.com' => json => {a => 'b'})->then(sub {
+    my $tx = shift;
+    say $tx->result->body;
+  })->catch(sub {
+    my $err = shift;
+    warn "Connection error: $err";
+  })->wait;
+
 =head2 patch
 
   my $tx = $ua->patch('example.com');
-  my $tx = $ua->patch('http://example.com' => {Accept => '*/*'} => 'Hi!');
+  my $tx = $ua->patch('http://example.com' => {Accept => '*/*'} => 'Content!');
   my $tx = $ua->patch(
     'http://example.com' => {Accept => '*/*'} => form => {a => 'b'});
   my $tx = $ua->patch(
@@ -799,14 +905,29 @@ implied). You can also append a callback to perform requests non-blocking.
 
   $ua->patch('http://example.com' => json => {a => 'b'} => sub {
     my ($ua, $tx) = @_;
-    say $tx->res->body;
+    say $tx->result->body;
   });
   Mojo::IOLoop->start unless Mojo::IOLoop->is_running;
 
+=head2 patch_p
+
+  my $promise = $ua->patch_p('http://example.com');
+
+Same as L</"patch">, but performs all requests non-blocking and returns a
+L<Mojo::Promise> object instead of accepting a callback.
+
+  $ua->patch_p('http://example.com' => json => {a => 'b'})->then(sub {
+    my $tx = shift;
+    say $tx->result->body;
+  })->catch(sub {
+    my $err = shift;
+    warn "Connection error: $err";
+  })->wait;
+
 =head2 post
 
   my $tx = $ua->post('example.com');
-  my $tx = $ua->post('http://example.com' => {Accept => '*/*'} => 'Hi!');
+  my $tx = $ua->post('http://example.com' => {Accept => '*/*'} => 'Content!');
   my $tx = $ua->post(
     'http://example.com' => {Accept => '*/*'} => form => {a => 'b'});
   my $tx = $ua->post(
@@ -819,14 +940,29 @@ implied). You can also append a callback to perform requests non-blocking.
 
   $ua->post('http://example.com' => json => {a => 'b'} => sub {
     my ($ua, $tx) = @_;
-    say $tx->res->body;
+    say $tx->result->body;
   });
   Mojo::IOLoop->start unless Mojo::IOLoop->is_running;
 
+=head2 post_p
+
+  my $promise = $ua->post_p('http://example.com');
+
+Same as L</"post">, but performs all requests non-blocking and returns a
+L<Mojo::Promise> object instead of accepting a callback.
+
+  $ua->post_p('http://example.com' => json => {a => 'b'})->then(sub {
+    my $tx = shift;
+    say $tx->result->body;
+  })->catch(sub {
+    my $err = shift;
+    warn "Connection error: $err";
+  })->wait;
+
 =head2 put
 
   my $tx = $ua->put('example.com');
-  my $tx = $ua->put('http://example.com' => {Accept => '*/*'} => 'Hi!');
+  my $tx = $ua->put('http://example.com' => {Accept => '*/*'} => 'Content!');
   my $tx = $ua->put(
     'http://example.com' => {Accept => '*/*'} => form => {a => 'b'});
   my $tx = $ua->put(
@@ -839,10 +975,25 @@ perform requests non-blocking.
 
   $ua->put('http://example.com' => json => {a => 'b'} => sub {
     my ($ua, $tx) = @_;
-    say $tx->res->body;
+    say $tx->result->body;
   });
   Mojo::IOLoop->start unless Mojo::IOLoop->is_running;
 
+=head2 put_p
+
+  my $promise = $ua->put_p('http://example.com');
+
+Same as L</"put">, but performs all requests non-blocking and returns a
+L<Mojo::Promise> object instead of accepting a callback.
+
+  $ua->put_p('http://example.com' => json => {a => 'b'})->then(sub {
+    my $tx = shift;
+    say $tx->result->body;
+  })->catch(sub {
+    my $err = shift;
+    warn "Connection error: $err";
+  })->wait;
+
 =head2 start
 
   my $tx = $ua->start(Mojo::Transaction::HTTP->new);
@@ -854,10 +1005,26 @@ to perform requests non-blocking.
   my $tx = $ua->build_tx(GET => 'http://example.com');
   $ua->start($tx => sub {
     my ($ua, $tx) = @_;
-    say $tx->res->body;
+    say $tx->result->body;
   });
   Mojo::IOLoop->start unless Mojo::IOLoop->is_running;
 
+=head2 start_p
+
+  my $promise = $ua->start_p(Mojo::Transaction::HTTP->new);
+
+Same as L</"start">, but performs all requests non-blocking and returns a
+L<Mojo::Promise> object instead of accepting a callback.
+
+  my $tx = $ua->build_tx(GET => 'http://example.com');
+  $ua->start_p($tx)->then(sub {
+    my $tx = shift;
+    say $tx->result->body;
+  })->catch(sub {
+    my $err = shift;
+    warn "Connection error: $err";
+  })->wait;
+
 =head2 websocket
 
   $ua->websocket('ws://example.com' => sub {...});
@@ -872,6 +1039,7 @@ object, depending on if the handshake was successful.
   $ua->websocket('wss://example.com/echo' => ['v1.proto'] => sub {
     my ($ua, $tx) = @_;
     say 'WebSocket handshake failed!' and return unless $tx->is_websocket;
+    say 'Subprotocol negotiation failed!' and return unless $tx->protocol;
     $tx->on(finish => sub {
       my ($tx, $code, $reason) = @_;
       say "WebSocket closed with status $code.";
@@ -887,10 +1055,34 @@ object, depending on if the handshake was successful.
 
 You can activate C<permessage-deflate> compression by setting the
 C<Sec-WebSocket-Extensions> header, this can result in much better performance,
-but also increases memory usage by up to 300KB per connection.
+but also increases memory usage by up to 300KiB per connection.
 
-  my $headers = {'Sec-WebSocket-Extensions' => 'permessage-deflate'};
-  $ua->websocket('ws://example.com/foo' => $headers => sub {...});
+  $ua->websocket('ws://example.com/foo' => {
+    'Sec-WebSocket-Extensions' => 'permessage-deflate'
+  } => sub {...});
+
+=head2 websocket_p
+
+  my $promise = $ua->websocket_p('ws://example.com');
+
+Same as L</"websocket">, but returns a L<Mojo::Promise> object instead of
+accepting a callback.
+
+  $ua->websocket_p('wss://example.com/echo')->then(sub {
+    my $tx = shift;
+    my $promise = Mojo::Promise->new;
+    $tx->on(finish => sub { $promise->resolve });
+    $tx->on(message => sub {
+      my ($tx, $msg) = @_;
+      say "WebSocket message: $msg";
+      $tx->finish;
+    });
+    $tx->send('Hi!');
+    return $promise;
+  })->catch(sub {
+    my $err = shift;
+    warn "WebSocket error: $err";
+  })->wait;
 
 =head1 DEBUGGING
 
@@ -901,6 +1093,6 @@ advanced diagnostics information printed to C<STDERR>.
 
 =head1 SEE ALSO
 
-L<Mojolicious>, L<Mojolicious::Guides>, L<http://mojolicio.us>.
+L<Mojolicious>, L<Mojolicious::Guides>, L<http://mojolicious.org>.
 
 =cut
