@@ -4,7 +4,10 @@ use Mojo::Base 'Mojo';
 # "Fry: Shut up and take my money!"
 use Carp ();
 use Mojo::Exception;
+use Mojo::Home;
+use Mojo::Log;
 use Mojo::Util;
+use Mojo::UserAgent;
 use Mojolicious::Commands;
 use Mojolicious::Controller;
 use Mojolicious::Plugins;
@@ -15,7 +18,6 @@ use Mojolicious::Static;
 use Mojolicious::Types;
 use Mojolicious::Validator;
 use Scalar::Util ();
-use Time::HiRes  ();
 
 has commands => sub {
   my $commands = Mojolicious::Commands->new(app => shift);
@@ -23,7 +25,22 @@ has commands => sub {
   return $commands;
 };
 has controller_class => 'Mojolicious::Controller';
-has mode => sub { $ENV{MOJO_MODE} || $ENV{PLACK_ENV} || 'development' };
+has home             => sub { Mojo::Home->new->detect(ref shift) };
+has log              => sub {
+  my $self = shift;
+
+  # Check if we have a log directory that is writable
+  my $log  = Mojo::Log->new;
+  my $home = $self->home;
+  my $mode = $self->mode;
+  $log->path($home->child('log', "$mode.log"))
+    if -d $home->child('log') && -w _;
+
+  # Reduced log output outside of development mode
+  return $mode eq 'development' ? $log : $log->level('info');
+};
+has 'max_request_size';
+has mode     => sub { $ENV{MOJO_MODE} || $ENV{PLACK_ENV} || 'development' };
 has moniker  => sub { Mojo::Util::decamelize ref shift };
 has plugins  => sub { Mojolicious::Plugins->new };
 has renderer => sub { Mojolicious::Renderer->new };
@@ -37,13 +54,18 @@ has secrets  => sub {
   # Default to moniker
   return [$self->moniker];
 };
-has sessions  => sub { Mojolicious::Sessions->new };
-has static    => sub { Mojolicious::Static->new };
-has types     => sub { Mojolicious::Types->new };
+has sessions => sub { Mojolicious::Sessions->new };
+has static   => sub { Mojolicious::Static->new };
+has types    => sub { Mojolicious::Types->new };
+has ua       => sub {
+  my $ua = Mojo::UserAgent->new;
+  Scalar::Util::weaken $ua->server->app(shift)->{app};
+  return $ua;
+};
 has validator => sub { Mojolicious::Validator->new };
 
-our $CODENAME = 'Clinking Beer Mugs';
-our $VERSION  = '6.24';
+our $CODENAME = 'Doughnut';
+our $VERSION  = '7.70';
 
 sub AUTOLOAD {
   my $self = shift;
@@ -78,8 +100,12 @@ sub build_controller {
 
 sub build_tx {
   my $self = shift;
-  my $tx   = Mojo::Transaction::HTTP->new;
+
+  my $tx  = Mojo::Transaction::HTTP->new;
+  my $max = $self->max_request_size;
+  $tx->req->max_message_size($max) if defined $max;
   $self->plugins->emit_hook(after_build_tx => $tx, $self);
+
   return $tx;
 }
 
@@ -102,7 +128,7 @@ sub dispatch {
     my $method = $req->method;
     my $path   = $req->url->path->to_abs_string;
     $self->log->debug(qq{$method "$path"});
-    $stash->{'mojo.started'} = [Time::HiRes::gettimeofday];
+    $c->helpers->timing->begin('mojo.timer');
   }
 
   # Routes
@@ -127,16 +153,10 @@ sub handler {
 
   # Delayed response
   $self->log->debug('Nothing has been rendered, expecting delayed response')
-    unless $c->tx->is_writing;
+    unless $c->stash->{'mojo.rendered'};
 }
 
-sub helper {
-  my ($self, $name, $cb) = @_;
-  my $r = $self->renderer;
-  $self->log->debug(qq{Helper "$name" already exists, replacing})
-    if exists $r->helpers->{$name};
-  $r->add_helper($name => $cb);
-}
+sub helper { shift->renderer->add_helper(@_) }
 
 sub hook { shift->plugins->on(@_) }
 
@@ -144,8 +164,8 @@ sub new {
   my $self = shift->SUPER::new(@_);
 
   my $home = $self->home;
-  push @{$self->renderer->paths}, $home->rel_dir('templates');
-  push @{$self->static->paths},   $home->rel_dir('public');
+  push @{$self->renderer->paths}, $home->child('templates')->to_string;
+  push @{$self->static->paths},   $home->child('public')->to_string;
 
   # Default to controller and application namespace
   my $r = $self->routes->namespaces(["@{[ref $self]}::Controller", ref $self]);
@@ -157,20 +177,12 @@ sub new {
   $r->hide(qw(rendered req res respond_to send session signed_cookie stash));
   $r->hide(qw(tx url_for validation write write_chunk));
 
-  # Check if we have a log directory that is writable
-  my $mode = $self->mode;
-  $self->log->path($home->rel_file("log/$mode.log"))
-    if -d $home->rel_file('log') && -w _;
-
   $self->plugin($_)
     for qw(HeaderCondition DefaultHelpers TagHelpers EPLRenderer EPRenderer);
 
   # Exception handling should be first in chain
   $self->hook(around_dispatch => \&_exception);
 
-  # Reduced log output outside of development mode
-  $self->log->level('info') unless $mode eq 'development';
-
   $self->startup;
 
   return $self;
@@ -183,7 +195,7 @@ sub plugin {
 
 sub start {
   my $self = shift;
-  $_->_warmup for $self->static, $self->renderer;
+  $_->warmup for $self->static, $self->renderer;
   return $self->commands->run(@_ ? @_ : @ARGV);
 }
 
@@ -192,7 +204,7 @@ sub startup { }
 sub _exception {
   my ($next, $c) = @_;
   local $SIG{__DIE__}
-    = sub { ref $_[0] ? CORE::die($_[0]) : Mojo::Exception->throw(@_) };
+    = sub { ref $_[0] ? CORE::die $_[0] : Mojo::Exception->throw(shift) };
   $c->helpers->reply->exception($@) unless eval { $next->(); 1 };
 }
 
@@ -228,6 +240,12 @@ Mojolicious - Real-time web framework
 
 =head1 DESCRIPTION
 
+An amazing real-time web framework built on top of the powerful L<Mojo> web
+development toolkit. With support for RESTful routes, plugins, commands,
+Perl-ish templates, content negotiation, session management, form validation,
+testing framework, static file server, C<CGI>/C<PSGI> detection, first class
+Unicode support and much more for you to discover.
+
 Take a look at our excellent documentation in L<Mojolicious::Guides>!
 
 =head1 HOOKS
@@ -249,6 +267,26 @@ rather advanced features such as upload progress bars possible. Note that this
 hook will not work for embedded applications, because only the host application
 gets to build transactions. (Passed the transaction and application object)
 
+=head2 around_dispatch
+
+Emitted right after a new request has been received and wraps around the whole
+dispatch process, so you have to manually forward to the next hook if you want
+to continue the chain. Default exception handling with
+L<Mojolicious::Plugin::DefaultHelpers/"reply-E<gt>exception"> is the first hook
+in the chain and a call to L</"dispatch"> the last, yours will be in between.
+
+  $app->hook(around_dispatch => sub {
+    my ($next, $c) = @_;
+    ...
+    $next->();
+    ...
+  });
+
+This is a very powerful hook and should not be used lightly, it allows you to,
+for example, customize application-wide exception handling, consider it the
+sledgehammer in your toolbox. (Passed a callback leading to the next hook and
+the default controller object)
+
 =head2 before_dispatch
 
 Emitted right before the static file server and router start their work.
@@ -289,7 +327,7 @@ controller object)
 
 =head2 around_action
 
-Emitted right before an action gets invoked and wraps around it, so you have to
+Emitted right before an action gets executed and wraps around it, so you have to
 manually forward to the next hook if you want to continue the chain. Default
 action dispatching is the last hook in the chain, yours will run before it.
 
@@ -301,9 +339,10 @@ action dispatching is the last hook in the chain, yours will run before it.
 
 This is a very powerful hook and should not be used lightly, it allows you for
 example to pass additional arguments to actions or handle return values
-differently. (Passed a callback leading to the next hook, the current
-controller object, the action callback and a flag indicating if this action is
-an endpoint)
+differently. Note that this hook can trigger more than once for the same
+request if there are nested routes. (Passed a callback leading to the next hook,
+the current controller object, the action callback and a flag indicating if this
+action is an endpoint)
 
 =head2 before_render
 
@@ -336,9 +375,9 @@ current controller object, a reference to the content and the format)
 
 =head2 after_dispatch
 
-Emitted in reverse order after a response has been rendered. Note that this
+Emitted in reverse order after a response has been generated. Note that this
 hook can trigger out of order due to its dynamic nature, and with embedded
-applications will only work for the application that is rendering.
+applications will only work for the application that is generating the response.
 
   $app->hook(after_dispatch => sub {
     my $c = shift;
@@ -348,26 +387,6 @@ applications will only work for the application that is rendering.
 Useful for rewriting outgoing responses and other post-processing tasks.
 (Passed the current controller object)
 
-=head2 around_dispatch
-
-Emitted right before the L</"before_dispatch"> hook and wraps around the whole
-dispatch process, so you have to manually forward to the next hook if you want
-to continue the chain. Default exception handling with
-L<Mojolicious::Plugin::DefaultHelpers/"reply-E<gt>exception"> is the first hook
-in the chain and a call to L</"dispatch"> the last, yours will be in between.
-
-  $app->hook(around_dispatch => sub {
-    my ($next, $c) = @_;
-    ...
-    $next->();
-    ...
-  });
-
-This is a very powerful hook and should not be used lightly, it allows you for
-example to customize application wide exception handling, consider it the
-sledgehammer in your toolbox. (Passed a callback leading to the next hook and
-the default controller object)
-
 =head1 ATTRIBUTES
 
 L<Mojolicious> inherits all attributes from L<Mojo> and implements the
@@ -393,16 +412,49 @@ Class to be used for the default controller, defaults to
 L<Mojolicious::Controller>. Note that this class needs to have already been
 loaded before the first request arrives.
 
+=head2 home
+
+  my $home = $app->home;
+  $app     = $app->home(Mojo::Home->new);
+
+The home directory of your application, defaults to a L<Mojo::Home> object
+which stringifies to the actual path.
+
+  # Portably generate path relative to home directory
+  my $path = $app->home->child('data', 'important.txt');
+
+=head2 log
+
+  my $log = $app->log;
+  $app    = $app->log(Mojo::Log->new);
+
+The logging layer of your application, defaults to a L<Mojo::Log> object. The
+level will default to C<debug> if the L</mode> is C<development>, or C<info>
+otherwise. All messages will be written to C<STDERR>, or a C<log/$mode.log> file
+if a C<log> directory exists.
+
+  # Log debug message
+  $app->log->debug('It works');
+
+=head2 max_request_size
+
+  my $max = $app->max_request_size;
+  $app    = $app->max_request_size(16777216);
+
+Maximum request size in bytes, defaults to the value of
+L<Mojo::Message/"max_message_size">. Setting the value to C<0> will allow
+requests of indefinite size. Note that increasing this value can also
+drastically increase memory usage, should you for example attempt to parse an
+excessively large request body with the methods L<Mojo::Message/"dom"> or
+L<Mojo::Message/"json">.
+
 =head2 mode
 
   my $mode = $app->mode;
   $app     = $app->mode('production');
 
 The operating mode for your application, defaults to a value from the
-C<MOJO_MODE> and C<PLACK_ENV> environment variables or C<development>. Right
-before calling L</"startup">, L<Mojolicious> will pick up the current mode,
-name the log file after it and raise the log level from C<debug> to C<info> if
-it has a value other than C<development>.
+C<MOJO_MODE> and C<PLACK_ENV> environment variables or C<development>.
 
 =head2 moniker
 
@@ -505,6 +557,9 @@ L<Mojolicious::Static> object.
   # Add another class with static files in DATA section
   push @{$app->static->classes}, 'Mojolicious::Plugin::Fun';
 
+  # Remove built-in favicon
+  delete $app->static->extra->{'favicon.ico'};
+
 =head2 types
 
   my $types = $app->types;
@@ -516,6 +571,17 @@ L<Mojolicious::Types> object.
   # Add custom MIME type
   $app->types->type(twt => 'text/tweet');
 
+=head2 ua
+
+  my $ua = $app->ua;
+  $app   = $app->ua(Mojo::UserAgent->new);
+
+A full featured HTTP user agent for use in your applications, defaults to a
+L<Mojo::UserAgent> object.
+
+  # Perform blocking request
+  say $app->ua->get('example.com')->result->body;
+
 =head2 validator
 
   my $validator = $app->validator;
@@ -529,6 +595,12 @@ Validate values, defaults to a L<Mojolicious::Validator> object.
     return $value ne 'foo';
   });
 
+  # Add validation filter
+  $app->validator->add_filter(quotemeta => sub {
+    my ($validation, $name, $value) = @_;
+    return quotemeta $value;
+  });
+
 =head1 METHODS
 
 L<Mojolicious> inherits all methods from L<Mojo> and implements the following
@@ -555,8 +627,8 @@ Build L<Mojo::Transaction::HTTP> object and emit L</"after_build_tx"> hook.
 
   my $hash = $app->defaults;
   my $foo  = $app->defaults('foo');
-  $app     = $app->defaults({foo => 'bar'});
-  $app     = $app->defaults(foo => 'bar');
+  $app     = $app->defaults({foo => 'bar', baz => 23});
+  $app     = $app->defaults(foo => 'bar', baz => 23);
 
 Default values for L<Mojolicious::Controller/"stash">, assigned for every new
 request.
@@ -587,8 +659,10 @@ every request.
 
   $app->helper(foo => sub {...});
 
-Add a new helper that will be available as a method of the controller object
-and the application object, as well as a function in C<ep> templates.
+Add or replace a helper that will be available as a method of the controller
+object and the application object, as well as a function in C<ep> templates. For
+a full list of helpers that are available by default see
+L<Mojolicious::Plugin::DefaultHelpers> and L<Mojolicious::Plugin::TagHelpers>.
 
   # Helper
   $app->helper(cache => sub { state $cache = {} });
@@ -626,10 +700,9 @@ requests indiscriminately, for a full list of available hooks see L</"HOOKS">.
   my $app = Mojolicious->new({moniker => 'foo_bar'});
 
 Construct a new L<Mojolicious> application and call L</"startup">. Will
-automatically detect your home directory and set up logging based on your
-current operating mode. Also sets up the renderer, static file server, a
-default set of plugins and an L</"around_dispatch"> hook with the default
-exception handling.
+automatically detect your home directory. Also sets up the renderer, static file
+server, a default set of plugins and an L</"around_dispatch"> hook with the
+default exception handling.
 
 =head2 plugin
 
@@ -651,9 +724,9 @@ L<Mojolicious> distribution see L<Mojolicious::Plugins/"PLUGINS">.
   $app->start;
   $app->start(@ARGV);
 
-Start the command line interface for your application, for a full list of
-commands available by default see L<Mojolicious::Commands/"COMMANDS">. Note
-that the options C<-h>/C<--help>, C<--home> and C<-m>/C<--mode>, which are
+Start the command line interface for your application. For a full list of
+commands that are available by default see L<Mojolicious::Commands/"COMMANDS">.
+Note that the options C<-h>/C<--help>, C<--home> and C<-m>/C<--mode>, which are
 shared by all commands, will be parsed from C<@ARGV> during compile time.
 
   # Always start daemon
@@ -693,20 +766,20 @@ that have been bundled for internal use.
 
 =head2 Mojolicious Artwork
 
-  Copyright (C) 2010-2015, Sebastian Riedel.
+  Copyright (C) 2010-2018, Sebastian Riedel.
 
 Licensed under the CC-SA License, Version 4.0
 L<http://creativecommons.org/licenses/by-sa/4.0>.
 
 =head2 jQuery
 
-  Copyright (C) 2005, 2014 jQuery Foundation, Inc.
+  Copyright (C) jQuery Foundation.
 
 Licensed under the MIT License, L<http://creativecommons.org/licenses/MIT>.
 
 =head2 prettify.js
 
-  Copyright (C) 2006, 2013 Google Inc.
+  Copyright (C) 2006, 2013 Google Inc..
 
 Licensed under the Apache License, Version 2.0
 L<http://www.apache.org/licenses/LICENSE-2.0>.
@@ -716,6 +789,8 @@ L<http://www.apache.org/licenses/LICENSE-2.0>.
 Every major release of L<Mojolicious> has a code name, these are the ones that
 have been used in the past.
 
+7.0, C<Doughnut> (U+1F369)
+
 6.0, C<Clinking Beer Mugs> (U+1F37B)
 
 5.0, C<Tiger Face> (U+1F42F)
@@ -726,20 +801,8 @@ have been used in the past.
 
 2.0, C<Leaf Fluttering In Wind> (U+1F343)
 
-1.4, C<Smiling Face With Sunglasses> (U+1F60E)
-
-1.3, C<Tropical Drink> (U+1F379)
-
-1.1, C<Smiling Cat Face With Heart-Shaped Eyes> (U+1F63B)
-
 1.0, C<Snowflake> (U+2744)
 
-0.999930, C<Hot Beverage> (U+2615)
-
-0.999927, C<Comet> (U+2604)
-
-0.999920, C<Snowman> (U+2603)
-
 =head1 SPONSORS
 
 Some of the work on this distribution has been sponsored by
@@ -755,12 +818,6 @@ Current members of the core team in alphabetical order:
 
 =over 2
 
-Abhijit Menon-Sen, C<ams@cpan.org>
-
-Dan Book, C<dbook@cpan.org>
-
-Glen Hinkle, C<tempire@cpan.org>
-
 Jan Henning Thorsen, C<jhthorsen@cpan.org>
 
 Joel Berger, C<jberger@cpan.org>
@@ -769,6 +826,16 @@ Marcus Ramberg, C<mramberg@cpan.org>
 
 =back
 
+The following members of the core team are currently on hiatus:
+
+=over 2
+
+Abhijit Menon-Sen, C<ams@cpan.org>
+
+Glen Hinkle, C<tempire@cpan.org>
+
+=back
+
 =head1 CREDITS
 
 In alphabetical order:
@@ -789,6 +856,8 @@ Alexey Likhatskiy
 
 Anatoly Sharifulin
 
+Andre Parker
+
 Andre Vieth
 
 Andreas Jaekel
@@ -797,6 +866,8 @@ Andreas Koenig
 
 Andrew Fresh
 
+Andrew Nugged
+
 Andrey Khozov
 
 Andrey Kuzmin
@@ -833,12 +904,16 @@ Charlie Brady
 
 Chas. J. Owens IV
 
+Chase Whitener
+
 Christian Hansen
 
 chromatic
 
 Curt Tilmes
 
+Dan Book
+
 Daniel Kimsey
 
 Danijel Tasov
@@ -859,10 +934,18 @@ Dominik Jarmulowicz
 
 Dominique Dumont
 
+Dotan Dimet
+
 Douglas Christopher Wilson
 
+Ettore Di Giacinto
+
+Eugen Konkov
+
 Eugene Toropov
 
+Flavio Poletti
+
 Gisle Aas
 
 Graham Barr
@@ -905,8 +988,12 @@ Kitamura Akatsuki
 
 Klaus S. Madsen
 
+Knut Arne Bjorndal
+
 Lars Balker Rasmussen
 
+Lee Johnson
+
 Leon Brocard
 
 Magnus Holm
@@ -955,6 +1042,8 @@ Pascal Gaudette
 
 Paul Evans
 
+Paul Robins
+
 Paul Tomlin
 
 Pavel Shaydo
@@ -983,8 +1072,14 @@ Robin Lee
 
 Roland Lammel
 
+Roy Storey
+
 Ryan Jendoubi
 
+Salvador Fandino
+
+Santiago Zarate
+
 Sascha Kiefer
 
 Scott Wiersdorf
@@ -1003,8 +1098,12 @@ Stanis Trendelenburg
 
 Steffen Ullrich
 
+Stephan Kulow
+
 Stephane Este-Gracias
 
+Stevan Little
+
 Steve Atkins
 
 Tatsuhiko Miyagawa
@@ -1015,6 +1114,8 @@ Tianon Gravi
 
 Tomas Znamenacek
 
+Tudor Constantin
+
 Ulrich Habel
 
 Ulrich Kautz
@@ -1029,6 +1130,8 @@ Viliam Pucik
 
 Wes Cravens
 
+William Lindley
+
 Yaroslav Korshak
 
 Yuki Kimoto
@@ -1041,7 +1144,7 @@ Zoffix Znet
 
 =head1 COPYRIGHT AND LICENSE
 
-Copyright (C) 2008-2015, Sebastian Riedel.
+Copyright (C) 2008-2018, Sebastian Riedel and others.
 
 This program is free software, you can redistribute it and/or modify it under
 the terms of the Artistic License version 2.0.
@@ -1049,6 +1152,6 @@ the terms of the Artistic License version 2.0.
 =head1 SEE ALSO
 
 L<https://github.com/kraih/mojo>, L<Mojolicious::Guides>,
-L<http://mojolicio.us>.
+L<http://mojolicious.org>.
 
 =cut
