@@ -8,7 +8,6 @@ use Mojo::URL;
 use Mojo::Util;
 use Mojolicious::Routes::Match;
 use Scalar::Util ();
-use Time::HiRes  ();
 
 has [qw(app tx)];
 has match =>
@@ -106,8 +105,8 @@ sub finish {
   my $self = shift;
 
   # WebSocket
-  my $tx = $self->tx;
-  $tx->finish(@_) and return $tx->is_established ? $self : $self->rendered(101)
+  my $tx = $self->tx || Carp::croak 'Connection already closed';
+  $tx->finish(@_) and return $tx->established ? $self : $self->rendered(101)
     if $tx->is_websocket;
 
   # Chunked stream
@@ -137,8 +136,8 @@ sub helpers { $_[0]->app->renderer->get_helper('')->($_[0]) }
 
 sub on {
   my ($self, $name, $cb) = @_;
-  my $tx = $self->tx;
-  $self->rendered(101) if $tx->is_websocket && !$tx->is_established;
+  my $tx = $self->tx || Carp::croak 'Connection already closed';
+  $self->rendered(101) if $tx->is_websocket && !$tx->established;
   return $tx->on($name => sub { shift; $self->$cb(@_) });
 }
 
@@ -155,7 +154,7 @@ sub redirect_to {
   # Don't override 3xx status
   my $res = $self->res;
   $res->headers->location($self->url_for(@_));
-  return $self->rendered($res->is_status_class(300) ? () : 302);
+  return $self->rendered($res->is_redirect ? () : 302);
 }
 
 sub render {
@@ -168,7 +167,7 @@ sub render {
   my $plugins = $app->plugins->emit_hook(before_render => $self, $args);
   my $maybe   = delete $args->{'mojo.maybe'};
 
-  my $ts = $args->{'mojo.to_string'};
+  my $ts = $args->{'mojo.string'};
   my ($output, $format) = $app->renderer->render($self, $args);
 
   # Maybe no 404
@@ -187,7 +186,7 @@ sub render_later { shift->stash('mojo.rendered' => 1) }
 
 sub render_maybe { shift->render(@_, 'mojo.maybe' => 1) }
 
-sub render_to_string { shift->render(@_, 'mojo.to_string' => 1) }
+sub render_to_string { shift->render(@_, 'mojo.string' => 1) }
 
 sub rendered {
   my ($self, $status) = @_;
@@ -201,11 +200,10 @@ sub rendered {
   if (!$stash->{'mojo.finished'} && ++$stash->{'mojo.finished'}) {
 
     # Disable auto rendering and stop timer
-    my $app = $self->render_later->app;
-    if (my $started = delete $stash->{'mojo.started'}) {
-      my $elapsed = sprintf '%f',
-        Time::HiRes::tv_interval($started, [Time::HiRes::gettimeofday()]);
-      my $rps  = $elapsed == 0 ? '??' : sprintf '%.3f', 1 / $elapsed;
+    my $app    = $self->render_later->app;
+    my $timing = $self->helpers->timing;
+    if (defined(my $elapsed = $timing->elapsed('mojo.timer'))) {
+      my $rps  = $timing->rps($elapsed) // '??';
       my $code = $res->code;
       my $msg  = $res->message || $res->default_message($code);
       $app->log->debug("$code $msg (${elapsed}s, $rps/s)");
@@ -218,8 +216,8 @@ sub rendered {
   return $self;
 }
 
-sub req { shift->tx->req }
-sub res { shift->tx->res }
+sub req { (shift->tx || Carp::croak 'Connection already closed')->req }
+sub res { (shift->tx || Carp::croak 'Connection already closed')->res }
 
 sub respond_to {
   my ($self, $args) = (shift, ref $_[0] ? $_[0] : {@_});
@@ -248,11 +246,11 @@ sub respond_to {
 
 sub send {
   my ($self, $msg, $cb) = @_;
-  my $tx = $self->tx;
+  my $tx = $self->tx || Carp::croak 'Connection already closed';
   Carp::croak 'No WebSocket connection to send message to'
     unless $tx->is_websocket;
   $tx->send($msg, $cb ? sub { shift; $self->$cb(@_) } : ());
-  return $tx->is_established ? $self : $self->rendered(101);
+  return $tx->established ? $self : $self->rendered(101);
 }
 
 sub session {
@@ -290,12 +288,11 @@ sub signed_cookie {
 sub stash { Mojo::Util::_stash(stash => @_) }
 
 sub url_for {
-  my $self = shift;
-  my $target = shift // '';
+  my ($self, $target) = (shift, shift // '');
 
   # Absolute URL
   return $target if Scalar::Util::blessed $target && $target->isa('Mojo::URL');
-  return Mojo::URL->new($target) if $target =~ m!^(?:[^:/?#]+:|//)!;
+  return Mojo::URL->new($target) if $target =~ m!^(?:[^:/?#]+:|//|#)!;
 
   # Base
   my $url  = Mojo::URL->new;
@@ -347,14 +344,12 @@ sub validation {
 
 sub write {
   my ($self, $chunk, $cb) = @_;
-  ($cb, $chunk) = ($chunk, undef) if ref $chunk eq 'CODE';
   $self->res->content->write($chunk, $cb ? sub { shift; $self->$cb(@_) } : ());
   return $self->rendered;
 }
 
 sub write_chunk {
   my ($self, $chunk, $cb) = @_;
-  ($cb, $chunk) = ($chunk, undef) if ref $chunk eq 'CODE';
   my $content = $self->res->content;
   $content->write_chunk($chunk, $cb ? sub { shift; $self->$cb(@_) } : ());
   return $self->rendered;
@@ -405,7 +400,7 @@ a L<Mojolicious> object.
   $c->app->log->debug('Hello Mojo');
 
   # Generate path
-  my $path = $c->app->home->rel_file('templates/foo/bar.html.ep');
+  my $path = $c->app->home->child('templates', 'foo', 'bar.html.ep');
 
 =head2 match
 
@@ -435,7 +430,7 @@ underlying connection might get closed early.
   my $address = $c->tx->remote_address;
   my $port    = $c->tx->remote_port;
 
-  # Increase size limit for WebSocket messages to 16MB
+  # Increase size limit for WebSocket messages to 16MiB
   $c->tx->max_websocket_size(16777216) if $c->tx->is_websocket;
 
   # Perform non-blocking operation without knowing the connection status
@@ -551,7 +546,7 @@ establish the WebSocket connection.
   # Do something after the transaction has been finished
   $c->on(finish => sub {
     my $c = shift;
-    $c->app->log->debug('We are done');
+    $c->app->log->debug('All data has been sent');
   });
 
   # Receive WebSocket message
@@ -577,7 +572,7 @@ establish the WebSocket connection.
 
   my $value = $c->param('foo');
   $c        = $c->param(foo => 'ba;r');
-  $c        = $c->param(foo => qw(ba;r baz));
+  $c        = $c->param(foo => 'ba;r', 'baz');
   $c        = $c->param(foo => ['ba;r', 'baz']);
 
 Access route placeholder values that are not reserved stash values, file
@@ -586,8 +581,8 @@ string and C<application/x-www-form-urlencoded> or C<multipart/form-data>
 message body, in that order. If there are multiple values sharing the same
 name, and you want to access more than just the last one, you can use
 L</"every_param">. Parts of the request body need to be loaded into memory to
-parse C<POST> parameters, so you have to make sure it is not excessively large,
-there's a 16MB limit by default.
+parse C<POST> parameters, so you have to make sure it is not excessively large.
+There's a 16MiB limit for requests by default.
 
   # Get first value
   my $first = $c->every_param('foo')->[0];
@@ -613,16 +608,21 @@ For more control you can also access request information directly.
   $c = $c->redirect_to('/index.html');
   $c = $c->redirect_to('http://example.com/index.html');
 
-Prepare a C<302> redirect response, takes the same arguments as L</"url_for">.
+Prepare a C<302> (if the status code is not already C<3xx>) redirect response
+with C<Location> header, takes the same arguments as L</"url_for">.
 
-  # Moved permanently
+  # Moved Permanently
   $c->res->code(301);
   $c->redirect_to('some_route');
 
+  # Temporary Redirect
+  $c->res->code(307);
+  $c->redirect_to('some_route');
+
 =head2 render
 
   my $bool = $c->render;
-  my $bool = $c->render(controller => 'foo', action => 'bar');
+  my $bool = $c->render(foo => 'bar', baz => 23);
   my $bool = $c->render(template => 'foo/index');
   my $bool = $c->render(template => 'index', format => 'html');
   my $bool = $c->render(data => $bytes);
@@ -681,7 +681,7 @@ automatic rendering would result in a response.
 =head2 render_maybe
 
   my $bool = $c->render_maybe;
-  my $bool = $c->render_maybe(controller => 'foo', action => 'bar');
+  my $bool = $c->render_maybe(foo => 'bar', baz => 23);
   my $bool = $c->render_maybe('foo/index', format => 'html');
 
 Try to render content, but do not call
@@ -726,17 +726,20 @@ Get L<Mojo::Message::Request> object from L</"tx">.
   my $req = $c->tx->req;
 
   # Extract request information
-  my $url   = $c->req->url->to_abs;
-  my $info  = $c->req->url->to_abs->userinfo;
-  my $host  = $c->req->url->to_abs->host;
-  my $agent = $c->req->headers->user_agent;
-  my $bytes = $c->req->body;
-  my $str   = $c->req->text;
-  my $hash  = $c->req->params->to_hash;
-  my $value = $c->req->json;
-  my $foo   = $c->req->json('/23/foo');
-  my $dom   = $c->req->dom;
-  my $bar   = $c->req->dom('div.bar')->first->text;
+  my $method = $c->req->method;
+  my $url    = $c->req->url->to_abs;
+  my $info   = $c->req->url->to_abs->userinfo;
+  my $host   = $c->req->url->to_abs->host;
+  my $agent  = $c->req->headers->user_agent;
+  my $custom = $c->req->headers->header('Custom-Header');
+  my $bytes  = $c->req->body;
+  my $str    = $c->req->text;
+  my $hash   = $c->req->params->to_hash;
+  my $all    = $c->req->uploads;
+  my $value  = $c->req->json;
+  my $foo    = $c->req->json('/23/foo');
+  my $dom    = $c->req->dom;
+  my $bar    = $c->req->dom('div.bar')->first->text;
 
 =head2 res
 
@@ -747,9 +750,12 @@ Get L<Mojo::Message::Response> object from L</"tx">.
   # Longer version
   my $res = $c->tx->res;
 
-  # Force file download by setting a custom response header
+  # Force file download by setting a response header
   $c->res->headers->content_disposition('attachment; filename=foo.png;');
 
+  # Use a custom response header
+  $c->res->headers->header('Custom-Header' => 'whatever');
+
   # Make sure response is cached correctly
   $c->res->headers->cache_control('public, max-age=300');
   $c->res->headers->append(Vary => 'Accept-Encoding');
@@ -762,14 +768,11 @@ Get L<Mojo::Message::Response> object from L</"tx">.
     any  => sub {...}
   );
 
-Automatically select best possible representation for resource from C<Accept>
-request header, C<format> stash value or C<format> C<GET>/C<POST> parameter,
+Automatically select best possible representation for resource from C<format>
+stash value, C<format> C<GET>/C<POST> parameter or C<Accept> request header,
 defaults to L<Mojolicious::Renderer/"default_format"> or rendering an empty
 C<204> response. Each representation can be handled with a callback or a hash
-reference containing arguments to be passed to L</"render">. Since browsers
-often don't really know what they actually want, unspecific C<Accept> request
-headers with more than one MIME type will be ignored, unless the
-C<X-Requested-With> header is set to the value C<XMLHttpRequest>.
+reference containing arguments to be passed to L</"render">.
 
   # Everything else than "json" and "xml" gets a 204 response
   $c->respond_to(
@@ -791,7 +794,7 @@ L<Mojolicious::Plugin::DefaultHelpers/"accepts">.
   $c = $c->send($chars => sub {...});
 
 Send message or frame non-blocking via WebSocket, the optional drain callback
-will be invoked once all data has been written. This method will automatically
+will be executed once all data has been written. This method will automatically
 respond to WebSocket handshake requests with a C<101> response status, to
 establish the WebSocket connection.
 
@@ -806,9 +809,10 @@ establish the WebSocket connection.
   $c->send({binary => encode_json({test => 'I ♥ Mojolicious!'})});
 
   # Send "Ping" frame
-  $c->send([1, 0, 0, 0, 9, 'Hello World!']);
+  use Mojo::WebSocket 'WS_PING';
+  $c->send([1, 0, 0, 0, WS_PING, 'Hello World!']);
 
-  # Make sure previous message has been written before continuing
+  # Make sure the first message has been written before continuing
   $c->send('First message!' => sub {
     my $c = shift;
     $c->send('Second message!');
@@ -831,7 +835,7 @@ usually defaults to C<15> seconds.
 Persistent data storage for the next few requests, all session data gets
 serialized with L<Mojo::JSON> and stored Base64 encoded in HMAC-SHA1 signed
 cookies, to prevent tampering. Note that cookies usually have a C<4096> byte
-(4KB) limit, depending on browser.
+(4KiB) limit, depending on browser.
 
   # Manipulate session
   $c->session->{foo} = 'bar';
@@ -863,8 +867,8 @@ failing signature verification will be automatically discarded.
 
   my $hash = $c->stash;
   my $foo  = $c->stash('foo');
-  $c       = $c->stash({foo => 'bar'});
-  $c       = $c->stash(foo => 'bar');
+  $c       = $c->stash({foo => 'bar', baz => 23});
+  $c       = $c->stash(foo => 'bar', baz => 23);
 
 Non-persistent data storage and exchange for the current request, application
 wide default values can be set with L<Mojolicious/"defaults">. Some stash
@@ -891,9 +895,22 @@ C<mojo.*> prefix are reserved for internal use.
   my $url = $c->url_for('//example.com/index.html');
   my $url = $c->url_for('http://example.com/index.html');
   my $url = $c->url_for('mailto:sri@example.com');
+  my $url = $c->url_for('#whatever');
 
 Generate a portable L<Mojo::URL> object with base for a path, URL or route.
 
+  # Rebuild URL for the current route
+  $c->url_for;
+
+  # Rebuild URL for the current route, but replace the "name" placeholder value
+  $c->url_for(name => 'sebastian');
+
+  # Absolute URL for the current route
+  $c->url_for->to_abs;
+
+  # Build URL for route "test" with two placeholder values
+  $c->url_for('test', name => 'sebastian', foo => 'bar');
+
   # "http://127.0.0.1:3000/index.html" if application was started with Morbo
   $c->url_for('/index.html')->to_abs;
 
@@ -921,11 +938,11 @@ validate file uploads as well as C<GET> and C<POST> parameters extracted from
 the query string and C<application/x-www-form-urlencoded> or
 C<multipart/form-data> message body. Parts of the request body need to be loaded
 into memory to parse C<POST> parameters, so you have to make sure it is not
-excessively large, there's a 16MB limit by default.
+excessively large. There's a 16MiB limit for requests by default.
 
   # Validate GET/POST parameter
   my $validation = $c->validation;
-  $validation->required('title')->size(3, 50);
+  $validation->required('title', 'trim')->size(3, 50);
   my $title = $validation->param('title');
 
   # Validate file upload
@@ -938,10 +955,9 @@ excessively large, there's a 16MB limit by default.
   $c = $c->write;
   $c = $c->write('');
   $c = $c->write($bytes);
-  $c = $c->write(sub {...});
   $c = $c->write($bytes => sub {...});
 
-Write dynamic content non-blocking, the optional drain callback will be invoked
+Write dynamic content non-blocking, the optional drain callback will be executed
 once all data has been written. Calling this method without a chunk of data
 will finalize the response headers and allow for dynamic content to be written
 later.
@@ -966,7 +982,6 @@ You can call L</"finish"> or write an empty chunk of data at any time to end
 the stream.
 
   HTTP/1.1 200 OK
-  Connection: keep-alive
   Date: Sat, 13 Sep 2014 16:48:29 GMT
   Content-Length: 6
   Server: Mojolicious (Perl)
@@ -992,13 +1007,12 @@ defaults to C<15> seconds.
   $c = $c->write_chunk;
   $c = $c->write_chunk('');
   $c = $c->write_chunk($bytes);
-  $c = $c->write_chunk(sub {...});
   $c = $c->write_chunk($bytes => sub {...});
 
-Write dynamic content non-blocking with C<chunked> transfer encoding, the
-optional drain callback will be invoked once all data has been written. Calling
-this method without a chunk of data will finalize the response headers and
-allow for dynamic content to be written later.
+Write dynamic content non-blocking with chunked transfer encoding, the optional
+drain callback will be executed once all data has been written. Calling this
+method without a chunk of data will finalize the response headers and allow for
+dynamic content to be written later.
 
   # Make sure previous chunk has been written before continuing
   $c->write_chunk('H' => sub {
@@ -1013,7 +1027,6 @@ You can call L</"finish"> or write an empty chunk of data at any time to end
 the stream.
 
   HTTP/1.1 200 OK
-  Connection: keep-alive
   Date: Sat, 13 Sep 2014 16:48:29 GMT
   Transfer-Encoding: chunked
   Server: Mojolicious (Perl)
@@ -1042,6 +1055,6 @@ L<Mojolicious::Plugin::TagHelpers>.
 
 =head1 SEE ALSO
 
-L<Mojolicious>, L<Mojolicious::Guides>, L<http://mojolicio.us>.
+L<Mojolicious>, L<Mojolicious::Guides>, L<http://mojolicious.org>.
 
 =cut
