@@ -4,8 +4,8 @@ use Mojo::Base 'Mojolicious::Plugin';
 use Mojo::Asset::File;
 use Mojo::ByteStream;
 use Mojo::DOM;
+use Mojo::File 'path';
 use Mojo::URL;
-use Mojo::Util 'slurp';
 use Pod::Simple::XHTML;
 use Pod::Simple::Search;
 
@@ -16,12 +16,8 @@ sub register {
   $app->renderer->add_handler(
     $conf->{name} || 'pod' => sub {
       my ($renderer, $c, $output, $options) = @_;
-
-      # Preprocess and render
-      my $handler = $renderer->handlers->{$preprocess};
-      return undef unless $handler->($renderer, $c, $output, $options);
-      $$output = _pod_to_html($$output);
-      return 1;
+      $renderer->handlers->{$preprocess}($renderer, $c, $output, $options);
+      $$output = _pod_to_html($$output) if defined $$output;
     }
   );
 
@@ -30,7 +26,7 @@ sub register {
 
   # Perldoc browser
   return undef if $conf->{no_perldoc};
-  my $defaults = {module => 'Mojolicious/Guides', format => 'html'};
+  my $defaults = {module => 'Mojolicious/Guides'};
   return $app->routes->any(
     '/perldoc/:module' => $defaults => [module => qr/[^.]+/] => \&_perldoc);
 }
@@ -51,8 +47,8 @@ sub _html {
 
   # Rewrite code blocks for syntax highlighting and correct indentation
   for my $e ($dom->find('pre > code')->each) {
-    my $str = $e->content;
-    next if $str =~ /^\s*(?:\$|Usage:)\s+/m || $str !~ /[\$\@\%]\w|-&gt;\w/m;
+    next if (my $str = $e->content) =~ /^\s*(?:\$|Usage:)\s+/m;
+    next unless $str =~ /[\$\@\%]\w|-&gt;\w|^use\s+\w/m;
     my $attrs = $e->attr;
     my $class = $attrs->{class};
     $attrs->{class} = defined $class ? "$class prettyprint" : 'prettyprint';
@@ -61,14 +57,13 @@ sub _html {
   # Rewrite headers
   my $toc = Mojo::URL->new->fragment('toc');
   my @parts;
-  for my $e ($dom->find('h1, h2, h3')->each) {
+  for my $e ($dom->find('h1, h2, h3, h4')->each) {
 
     push @parts, [] if $e->tag eq 'h1' || !@parts;
-    my $anchor = $e->{id};
-    my $link   = Mojo::URL->new->fragment($anchor);
+    my $link = Mojo::URL->new->fragment($e->{id});
     push @{$parts[-1]}, my $text = $e->all_text, $link;
     my $permalink = $c->link_to('#' => $link, class => 'permalink');
-    $e->content($permalink . $c->link_to($text => $toc, id => $anchor));
+    $e->content($permalink . $c->link_to($text => $toc));
   }
 
   # Try to find a title
@@ -77,8 +72,7 @@ sub _html {
 
   # Combine everything to a proper response
   $c->content_for(perldoc => "$dom");
-  my $template = $c->app->renderer->_bundled('perldoc');
-  $c->render(inline => $template, title => $title, parts => \@parts);
+  $c->render('mojo/perldoc', title => $title, parts => \@parts);
 }
 
 sub _perldoc {
@@ -86,12 +80,12 @@ sub _perldoc {
 
   # Find module or redirect to CPAN
   my $module = join '::', split('/', $c->param('module'));
+  $c->stash(cpan => "https://metacpan.org/pod/$module");
   my $path
     = Pod::Simple::Search->new->find($module, map { $_, "$_/pods" } @INC);
-  return $c->redirect_to("https://metacpan.org/pod/$module")
-    unless $path && -r $path;
+  return $c->redirect_to($c->stash('cpan')) unless $path && -r $path;
 
-  my $src = slurp $path;
+  my $src = path($path)->slurp;
   $c->respond_to(txt => {data => $src}, html => sub { _html($c, $src) });
 }
 
@@ -198,6 +192,6 @@ Register renderer and helper in L<Mojolicious> application.
 
 =head1 SEE ALSO
 
-L<Mojolicious>, L<Mojolicious::Guides>, L<http://mojolicio.us>.
+L<Mojolicious>, L<Mojolicious::Guides>, L<http://mojolicious.org>.
 
 =cut
